{{
 "DioReadWriter", Initiates and sets up a fre-running reader/writer of remote GPIO, 1.0, works with mcp23017 only, so far
 Erlend Fjosna 201
-----------------------------------------------------------------------------------------------------------------------------------------------------------------------
 Init is a method to configure the GPIO for input/output, and w/wo pull-up resistors
 Start/Stop controls a method in separate cog which continously writes to and reads from the IO banks of the GPIO, using referenced global variables from parent object
-----------------------------------------------------------------------------------------------------------------------------------------------------------------------
}}
{
 Acknowledgements:
 I have learnt from JonnyMac's code as ususal>.

 Naming convention: Variables starts with:     stn: String, int: Integer, flt: Floating point, cmp: Composite, chr: Character, dta: Consecutive bytes containing data sets
                    Constant starts with:      Cstn: String, Cint: Integer, Cflt: Floating point, Ccmp: Composite, Cchr: Character, Cdta: Consecutive bytes containing data sets
                    DAT starts with:           Dstn: String, Dint: Integer, Dflt: Floating point, Dcmp: Composite, Dchr: Character, Ddta: Consecutive bytes containing data sets
                    PIN constant numbers are: 'intFncPIN' wher Fnc is an abbreviation of the signal/function
                    Pointers start with ptr followed by the variable core name, e.g. ptrVolume
                    Reserved words are always in capitals
                    Included objects are called by three or four letter lower case names
                    Methods (PUB and PRI) have first letter capital then lower case descriptive.
                    Local variables are free text beginning in lower case, or if they are copies begin with capitol L 

=======================================================================================================================================================================


=======================================================================================================================================================================
}

CON

          _clkmode = xtal1 + pll16x
          _xinfreq = 5_000_000                                          ' use 5MHz crystal
        
          clk_freq = (_clkmode >> 6) * _xinfreq                         ' system freq as a constant
          mSec     = clk_freq / 1_000                                   ' ticks in 1ms
          uSec     = clk_freq / 1_000_000                               ' ticks in 1us
        
                
VAR

           LONG  RWstack[128]
           BYTE  cog '
        

OBJ
           dio      : "MCP23017"


PUB Init(slcPin, sdaPin, chipAddress, ioDirA, ioDirB, puResA, puResB)                            'Start and configure MCP23017 with io direction and pullup resistors

   dio.start(slcPin, sdaPin)
   dio.cfg_dira(ioDirA, chipAddress)                                                             'Set i/o direction of bank A, device address, %000 to %111
   dio.cfg_dirb(ioDirB, chipAddress)                                                             'Set i/o direction of bank B, 0=input, 1=output
   dio.cfg_pua(puResA, chipAddress)                                                              'Switch on/off internal pullup resistors, bank A
   dio.cfg_pub(puResB, chipAddress)                                                              'Switch on/off internal pullup resistors, bank B
   

PUB Start(slcPin, sdaPin, chipAddress, ptrOutA, ptrOutB, ptrInA, ptrInB): Success

   Stop
   
   Success:= (cog:= COGNEW(ContinousRWdio(slcPin, sdaPin, chipAddress, ptrOutA, ptrOutB, ptrInA, ptrInB), @RWstack) + 1)    'Start continous read&write method


PUB Stop
   IF Cog
      COGSTOP(cog~ - 1)

      
PRI ContinousRWdio(slcPin, sdaPin, chipAddress, ptrOutA, ptrOutB, ptrInA, ptrInB)      'Writes to io configured as inputs are ignored by mcp23017

   REPEAT                                                                                         'Forever:
     dio.wr_gpioa(LONG[ptrOutA], chipAddress)                                                     'Write OUtA value to bank A
     LONG[ptrInA]:= dio.rd_gpioa(chipAddress)                                                     'Read into InA the value of bank A
     dio.wr_gpiob(LONG[ptrOutB], chipAddress)                                                     'Write OUtB value to bank B       
     LONG[ptrInB]:= dio.rd_gpiob(chipAddress)                                                     'Read into InB the value of bank B

     
DAT
           DaaaBbbbb    BYTE  "string_data",0        

{{

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and    
associated documentation files (the "Software"), to deal in the Software without restriction,        
including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:                                                                 
                                                                                                     
The above copyright notice and this permission notice shall be included in all copies or substantial 
portions of the Software.                                                                            
                                                                                                     
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER         
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

}}                                    